﻿/* Author: Jim-Kyrre Benjaminsen AKA Romfarer
 * License: This work is licensed under a Creative Commons Attribution-NonCommercial-NoDerivs 3.0 Unported License. 
 * http://creativecommons.org/licenses/by-nc-nd/3.0/
*/

using System;
using System.Linq;
using System.Collections.Generic;
using UnityEngine;
using KSP.IO;


namespace Romfarer {
    class GUIachievements : GUItab {

        /* File format
         * SOI # Type # Date # MissionTime # VesselName
         * Mun#Landed#12.11.2012:09:30#LazorCraft
         * Mun#Orbit#12.11.2012:09:30#LazorCraft
         * 
         * return missions
         * + land near easter eggs
         * crash into sun
         * orbit with eva
         * */

        /*counter exception handling
         * ctr f10                  !Application.isLoadingLevel
         * save/load                !Application.isLoadingLevel
         * close game               OnApplicationQuit()
         * end flight ,             !FlightDriver.Pause
         * eva beam aboard          != FlightGlobals.ActiveVessel in DestructionTrackerEVA, trigger with vessel.OnJustAboutToBeDestroyed in LazorEVA
         * eva beam target aboard   send teleportLock to DestructionTrackerEVA
         * eva board                
         * */
        
        private int achievements_Possible = 0;
        private int achievements_CompletedTotal = 0;
        private Dictionary<string, int> achievements_Completed = new Dictionary<string, int>();
        private Dictionary<string, List<Achievement>> achievements_All = new Dictionary<string, List<Achievement>>();
        private HashSet<string> collapsedItems = new HashSet<string>();
        private bool checkLanded;
        private bool checkReturn;
        private bool checkReturnComplete;
        private bool checkOrbit;
        private bool checkGeoOrbit;
        private bool checkEscape;
        private bool checkReachSpace;
        private HashSet<string> returnList = new HashSet<string>(); //list of bodies you have landed on, which will be "Return" completed when landed at kerbin
        private float geosyncAlt; //target altitude for geostationary orbit achievement
        private static char[] delimiters = new char[] { '#', '\t' };
        private int counter_KerbalsKilled = 0;
        private int counter_VesselsDestroyed = 0;

        private bool toggleSoI;
        private bool toggleCompleted = true;
        private Vector2 scrollArea = Vector2.zero;
        private Rect area;
        private GUIStyle listItem;
        private GUIStyle listItemCompleted;
        private GUIStyle listItemHeader;
        private Texture2D listItemBG = Util.getTexture(new Color(0.1f, 0.1f, 0.1f, 0.9f));
        private Texture2D listItemBGHeader = Util.getTexture(new Color(0.2f, 0.2f, 0.2f, 0.9f));
		private Color colorHeader = LazorGlobals.COLOR_GREENISH;

        public enum AchievementType {
            Landed = 1, // not kerbin
            Return = 6, // not kerbin

            Orbit = 2,
            GeoOrbit = 3,   //only for planets, gilly, minmus and bop
            Escape = 4, //sun and kerbin only

            ReachSpace = 5, //kerbin only
        }

        internal class Achievement{
            public string SoI;
            public AchievementType aT;
            public System.DateTime time;
            public double missionTime;
            public string vesselName;
            public bool completed;
            public Achievement(string SoI, AchievementType aT) {
                this.SoI = SoI;
                this.aT = aT;
            }
            public Achievement(string SoI, AchievementType aT, System.DateTime time, double missionTime, string vesselName) {
                this.SoI = SoI;
                this.aT = aT;
                this.time = time;
                this.missionTime = missionTime;
                this.vesselName = vesselName;
            }
        }

		override public void OnDestroy() { 
			system.localEvents.onSoIChange -= OnSoIChange;
			system.localEvents.onSituationChange -= OnSituationChange;
			Events.onKerbalKilled -= OnKerbalKilled;
			Events.onVesselDestroyed -= OnVesselDestroyed;
		}

        public GUIachievements(LazorSystem system) : base(system) {
			system.localEvents.onSoIChange += OnSoIChange;
            system.localEvents.onSituationChange += OnSituationChange;
            Events.onKerbalKilled += OnKerbalKilled;
            Events.onVesselDestroyed += OnVesselDestroyed;

            //Add all possible achievements to achievements_All
            foreach (CelestialBody cb in FlightGlobals.Bodies) {
                if (!achievements_All.ContainsKey(cb.name)) {
                    achievements_All.Add(cb.name, new List<Achievement>());
                    achievements_Completed.Add(cb.name, 0);
                }
                if (cb.name != "Kerbin" && cb.name != "Sun") {
                    achievements_All[cb.name].Add(new Achievement(cb.name, AchievementType.Landed));
                    achievements_Possible++;
                    achievements_All[cb.name].Add(new Achievement(cb.name, AchievementType.Return));
                    achievements_Possible++;
                }
                achievements_All[cb.name].Add(new Achievement(cb.name, AchievementType.Orbit));
                achievements_Possible++;
                if (cb.name == "Sun" || cb.name == "Kerbin" || cb.name == "Eve" || cb.name == "Duna" || cb.name == "Moho" || cb.name == "Jool" || cb.name == "Gilly" || cb.name == "Minmus" || cb.name == "Bop") {
                    achievements_All[cb.name].Add(new Achievement(cb.name, AchievementType.GeoOrbit));
                    achievements_Possible++;
                }
                if (cb.name == "Kerbin") {
                    achievements_All[cb.name].Add(new Achievement(cb.name, AchievementType.ReachSpace));
                    achievements_Possible++;
                }
                if (cb.name == "Kerbin" || cb.name == "Sun") {
                    achievements_All[cb.name].Add(new Achievement(cb.name, AchievementType.Escape));
                    achievements_Possible++;
                }
            }
            Load();
            LoadCounters();
        }

        override public string Name { get { return "Achievements"; } }
        override public string GetTabName() {
            return "Achievements";
        }

        public void OnSoIChange() { //register pending achievements for new SoI and initiate checks
            if (achievements_All.ContainsKey(system.vessel.mainBody.name)) {
                checkLanded = false;
                checkOrbit = false;
                checkGeoOrbit = false;
                checkEscape = false;
                checkReachSpace = false;
                checkReturn = false;
                checkReturnComplete = false;
                foreach(Achievement a in achievements_All[system.vessel.mainBody.name].FindAll(x => !x.completed)){
                    if ((int)a.aT == (int)AchievementType.Landed)
                        checkLanded = true;
                    else if ((int)a.aT == (int)AchievementType.Return)
                        checkReturn = true;
                    else if ((int)a.aT == (int)AchievementType.Orbit)
                        checkOrbit = true;
                    else if ((int)a.aT == (int)AchievementType.GeoOrbit) {
                        checkGeoOrbit = true;
                        geosyncAlt = LazorGlobals.GeosyncAltVel(system.vessel.mainBody.name).x;
                        //geosyncVel = LazorGlobals.GeosyncAltVel(system.vessel.mainBody.name).y;
                    } else if ((int)a.aT == (int)AchievementType.Escape)
                        checkEscape = true;
                    else if ((int)a.aT == (int)AchievementType.ReachSpace)
                        checkReachSpace = true;
                }
                //check return list for errors that can occur when loading saved stuff after actually completing the return achievement
                returnList.RemoveWhere(x => achievements_All.Keys.Contains(x) && achievements_All[x].Exists(y => y.completed && y.SoI == x && (int)y.aT == (int)AchievementType.Return));
                
                if (system.vessel.mainBody.name == "Kerbin" && returnList.Count > 0)
                    checkReturnComplete = true;
            }
        }

        public void OnSituationChange() {
            //check for landed achievements
            if ((int)system.vessel.situation == (int)Vessel.Situations.LANDED) { //ACHIEVEMENT TRIGGER
                if (checkLanded) {
                    CompleteAchievement(system.vessel.mainBody.name, AchievementType.Landed, "[" + DateTime.Now.ToString() + " Achievement completed: Landed on " + system.vessel.mainBody.name + " with the vessel " + system.vessel.vesselName + "]");
                    checkLanded = false;
                } 
                if (checkReturn) {
                    returnList.Add(system.vessel.mainBody.name);
                    checkLanded = false;
                }
                if (checkReturnComplete) {
                    foreach (string s in returnList) {
                        CompleteAchievement(s, AchievementType.Return, "[" + DateTime.Now.ToString() + " Achievement completed: Returned from " + s + " with the vessel " + system.vessel.vesselName + "]");
                    }
                    returnList.Clear();
                    checkReturnComplete = false;
                }
            }
            if (checkEscape) {
                if ((int)system.vessel.situation == (int)Vessel.Situations.ESCAPING) { //ACHIEVEMENT TRIGGER
                    CompleteAchievement(system.vessel.mainBody.name, AchievementType.Escape, "[" + DateTime.Now.ToString() + " Achievement completed: Escape orbit registered from " + system.vessel.mainBody.name + " with the vessel " + system.vessel.vesselName + "]");
                    checkEscape = false;
                }
            }
        }

        public void OnKerbalKilled() {
            counter_KerbalsKilled++;
            SaveCounters();
        }
        public void OnVesselDestroyed() {
            counter_VesselsDestroyed++;
            SaveCounters();
        }

        override public void OnFixedUpdate() {
            if (checkReachSpace) {
                if (system.vessel.altitude > 70000) { //ACHIEVEMENT TRIGGER
                    CompleteAchievement(system.vessel.mainBody.name, AchievementType.ReachSpace, "[" + DateTime.Now.ToString() + " Achievement completed: Reached space with the vessel " + system.vessel.vesselName + "]");
                    checkReachSpace = false;
                }
            }
            if (checkOrbit) {
                if ((int)system.vessel.situation == (int)Vessel.Situations.ORBITING) { //ACHIEVEMENT TRIGGER
                    if ((system.vessel.mainBody.atmosphere && system.vessel.orbit.PeA > system.vessel.mainBody.maxAtmosphereAltitude)
                        || (!system.vessel.mainBody.atmosphere && system.vessel.orbit.PeA > 1000)) {
                        CompleteAchievement(system.vessel.mainBody.name, AchievementType.Orbit, "[" + DateTime.Now.ToString() + " Achievement completed: Orbit recorded around " + system.vessel.mainBody.name + " with the vessel " + system.vessel.vesselName + "]");
                        checkOrbit = false;
                    }
                }
            }
            if (checkGeoOrbit) {
                if ((int)system.vessel.situation == (int)Vessel.Situations.ORBITING) { //ACHIEVEMENT TRIGGER
                    if (Util.CloseEnough(system.vessel.orbit.ApA, geosyncAlt, 1f) && Util.CloseEnough(system.vessel.orbit.PeA, geosyncAlt, 1f)) {
                        CompleteAchievement(system.vessel.mainBody.name, AchievementType.GeoOrbit, "[" + DateTime.Now.ToString() + " Achievement completed: Geostationary orbit recorded around " + system.vessel.mainBody.name + " with the vessel " + system.vessel.vesselName + "]");
                        checkGeoOrbit = false;
                    }
                }
            }
        }

        private void CompleteAchievement(string SoI, AchievementType type, string message) {
            MonoBehaviour.print(message);
            Achievement a = achievements_All[SoI].Find(x => (int)x.aT == (int)type);
            a.completed = true;
            a.time = DateTime.Now;
            a.missionTime = system.vessel.missionTime;
            a.vesselName = system.vessel.vesselName;
            achievements_Completed[a.SoI]++;
            achievements_CompletedTotal++;
            system.lazorSound.Play(SoundType.ACHIEVEMENT);
            Save();
        }

        public int Load() { //load achievement list
            if (!File.Exists<GUIachievements>("achievements.txt")) {
                Save();
            } else {
                try {
                    TextReader tr = TextReader.CreateForType<GUIachievements>("achievements.txt");
                    while (tr.Peek() != -1) {
                        string l = tr.ReadLine();
                        string[] item = l.Split(delimiters);
                        Achievement a = achievements_All[item[0]].Find(x => (int)x.aT == int.Parse(item[1]));
                        a.completed = true;
                        a.time = System.DateTime.Parse(item[2]);
                        a.missionTime = Double.Parse(item[3]);
                        a.vesselName = item[4];
                        achievements_Completed[a.SoI]++;
                        achievements_CompletedTotal++;
                    }
                    tr.Close();
                    tr.Dispose();
                } catch (IOException) { } catch (Exception) { }
            }
            return 0;
        }
        public void Save() {
            //save achievement list
            try {
                TextWriter tw = TextWriter.CreateForType<GUIachievements>("achievements.txt");
                foreach(List<Achievement> l in achievements_All.Values)
                    foreach(Achievement a in l)
                        if (a.completed) {
                            tw.WriteLine(a.SoI + "#" + (int)a.aT + "#" + a.time.ToString() + "#" + a.missionTime.ToString() + "#" + a.vesselName);
                        }
                tw.Flush();
                tw.Close();
                tw.Dispose();

            } catch (IOException) { } catch (Exception) { }
        }

        public void LoadCounters() { //load achievement list
            if (!File.Exists<GUIachievements>("counters.txt")) {
                SaveCounters();
            } else {
                try {
                    TextReader tr = TextReader.CreateForType<GUIachievements>("counters.txt");
                    while (tr.Peek() != -1) {
                        string l = tr.ReadLine();
                        string[] item = l.Split(delimiters);
                        counter_KerbalsKilled = int.Parse(item[0]);
                        counter_VesselsDestroyed = int.Parse(item[1]);
                    }
                    tr.Close();
                    tr.Dispose();
                } catch (IOException) { } catch (Exception) { }
            }
        }
        public void SaveCounters() {
            try {
                TextWriter tw = TextWriter.CreateForType<GUIachievements>("counters.txt");
                tw.WriteLine(counter_KerbalsKilled + "#" + counter_VesselsDestroyed);
                tw.Flush();
                tw.Close();
                tw.Dispose();
            } catch (IOException) { } catch (Exception) { }
        }




        //*******************************************************************************************************************
        //*******************************************************************************************************************
        // GUI
        //*******************************************************************************************************************
        //*******************************************************************************************************************


        override public void DrawGUI(string name) {
            GUI.skin = Util.LazorSkin;
            windowPos = GUILayout.Window(id, windowPos, WindowGUI, name, Util.Window_border);
            if (!docked)
                windowPos = Util.ClampToScreen(windowPos);
        }

        override public void WindowGUI(int windowID) {
            Header();
            if (listItem == null) {
                listItem = new GUIStyle(Util.ListItem_std);
                listItem.normal.background = listItemBG;
                listItem.hover.background = listItemBG;
            }
            if (listItemCompleted == null) {
                listItemCompleted = new GUIStyle(Util.ListItem_std);
                listItemCompleted.normal.background = listItemBG;
                listItemCompleted.hover.background = listItemBG;
                listItemCompleted.normal.textColor = new Color(0.6f, 0.6f, 0.6f, 1);
            }
            if (listItemHeader == null) {
                listItemHeader = new GUIStyle(Util.ListItem_std);
                listItemHeader.normal.background = listItemBGHeader;
                listItemHeader.hover.background = listItemBGHeader;
                listItemHeader.normal.textColor = colorHeader;
            }

            GUILayout.BeginVertical(Util.Space_Formatted);

            GUILayout.BeginHorizontal();
            GUILayout.Label("Completed: ", Util.StdLabelSmallCustom(colorHeader));
            GUILayout.Label(achievements_CompletedTotal + " / " + achievements_Possible, Util.StdLabelSmallCustom(Color.white));
            GUILayout.EndHorizontal();

            GUILayout.BeginHorizontal();
            GUILayout.Label("Kerbals Killed:", Util.StdLabelSmallCustom(colorHeader));
            GUILayout.Label(counter_KerbalsKilled.ToString(), Util.StdLabelSmallCustom(Color.white));
            GUILayout.EndHorizontal();

            GUILayout.BeginHorizontal();
            GUILayout.Label("Vessels Destroyed:", Util.StdLabelSmallCustom(colorHeader));
            GUILayout.Label(counter_VesselsDestroyed.ToString(), Util.StdLabelSmallCustom(Color.white));
            GUILayout.EndHorizontal();

            GUILayout.BeginHorizontal();
            toggleSoI = GUILayout.Toggle(toggleSoI, "SoI Only", Util.Button_Toggle);
            toggleCompleted = GUILayout.Toggle(toggleCompleted, "Show Completed", Util.Button_Toggle);
            GUILayout.EndHorizontal();

            GUILayout.EndVertical();

            float height = area.height > 400 ? 410 : area.height + 10;
            scrollArea = GUILayout.BeginScrollView(scrollArea, false, false, GUILayout.MinHeight(height), GUILayout.ExpandHeight(false));//, Util.TxtAreaBars, Util.TxtAreaBars, Util.TxtArea, GUILayout.MinHeight(height), GUILayout.ExpandHeight(false));
            GUILayout.BeginVertical(GUILayout.MaxWidth(windowPos.width - 22));

            foreach (string s in achievements_All.Keys) {
                if (toggleSoI) {
                    if (s == system.vessel.mainBody.name) {
                        GUILayout.Label(s + " (" + achievements_Completed[s] + "/" + achievements_All[s].Count + ")", listItemHeader);
                        foreach (Achievement a in achievements_All[s]) {
                            if (a.completed) {
                                if (toggleCompleted)
                                    GUILayout.Label("   " + a.aT.ToString() + " " + a.time.ToShortDateString(), listItemCompleted);
                            } else
                                GUILayout.Label("   " + a.aT.ToString() + ((a.aT == AchievementType.Return && returnList.Contains(s)) ? " (R)" : ""), listItem);
                        }
                    }
                } else {
                    if (collapsedItems.Contains(s)) {
                        if (GUILayout.Button(s + " (" + achievements_Completed[s] + "/" + achievements_All[s].Count + ")", listItemHeader))
                            collapsedItems.Remove(s);
                    } else {
                        if (GUILayout.Button(s + " (" + achievements_Completed[s] + "/" + achievements_All[s].Count + ")", listItemHeader))
                            collapsedItems.Add(s);
                        foreach (Achievement a in achievements_All[s]) {
                            if (a.completed) {
                                if (toggleCompleted)
                                    GUILayout.Label("   " + a.aT.ToString() + " " + a.time.ToShortDateString(), listItemCompleted);
                            } else {
                                GUILayout.Label("   " + a.aT.ToString() + ((a.aT == AchievementType.Return && returnList.Contains(s)) ? " (progress)" : ""), listItem);
                            }
                        }
                    }
                }
            }
            GUILayout.EndVertical();
            if (Event.current.type == EventType.Repaint)
                area = GUILayoutUtility.GetLastRect();
            GUILayout.EndScrollView();

            Footer();
        }


        public void onFlightStateLoad(Dictionary<string, KSPParseable> parsedData) {
            //load inprogress achievements
            if (parsedData != null)
                if (parsedData.ContainsKey("returnList"))
                    returnList.UnionWith(parsedData["returnList"].value.Split(delimiters));
            returnList.Remove(""); //something about trailing '#' in the list format made by string.join that causes the string.split to add empty strings.
            onFlightStateLoad(parsedData, "GUIachievements"); //load generic stuff from base class
        }
        public void onFlightStateSave(Dictionary<string, KSPParseable> partDataCollection) {
            //save in progress achievements
            partDataCollection.Add("returnList", new KSPParseable(string.Join("#", returnList.ToArray<string>()), KSPParseable.Type.STRING));
            onFlightStateSave(partDataCollection, "GUIachievements");
        }
    }
}
